from tuikit.core.signal import Signal
from tuikit.core.coords import Size, Rect
from tuikit.core.theme import default_theme

from contextlib import contextmanager


class Cell:

    """Character buffer cell.

    Cell contains one character and its attribute.

    """

    __slots__ = ('char', 'attr')

    def __init__(self, char=' ', attr=0):
        self.char = char
        self.attr = attr

    def set(self, char, attr):
        self.char = char
        self.attr = attr


class BufferOperationsMixin:

    def draw(self, buffer, x=0, y=0):
        """Draw another buffer onto this buffer at x/y coords."""
        for bufy in range(buffer.size.h):
            for bufx in range(buffer.size.w):
                char, attr_desc = buffer.get(bufx, bufy)
                self.setattr(attr_desc)
                self.putch(x + bufx, y + bufy, char)

    def puts(self, x, y, s):
        """Output string of characters."""
        for c in s:
            self.putch(x, y, c)
            x += 1

    def hline(self, x, y, w, c=' '):
        """Draw horizontal line."""
        self.puts(x, y, [c] * w)

    def vline(self, x, y, h, c=' '):
        """Draw vertical line."""
        for i in range(h):
            self.putch(x, y + i, c)

    def fill(self, x=0, y=0, w=0, h=0, c=' '):
        """Fill rectangular area.

        Fill whole buffer if width or height is not specified (zero).

        """
        w = self.size.w if not w else w
        h = self.size.h if not h else h
        for i in range(h):
            self.hline(x, y + i, w, c)

    def frame(self, x=0, y=0, w=0, h=0, theme=default_theme):
        """Draw rectangular frame.

        Frame whole buffer if width or height is not specified (zero).
        Use line-drawing characters from `style` bank.

        """
        w = self.size.w if not w else w
        h = self.size.h if not h else h
        self.putch(x,      y,      theme.frame_ulcorner)
        self.putch(x+w-1,  y,      theme.frame_urcorner)
        self.putch(x,      y+h-1,  theme.frame_llcorner)
        self.putch(x+w-1,  y+h-1,  theme.frame_lrcorner)
        self.hline(x+1,    y,      w-2,  theme.frame_hline)
        self.hline(x+1,    y+h-1,  w-2,  theme.frame_hline)
        self.vline(x,      y+1,    h-2,  theme.frame_vline)
        self.vline(x+w-1,  y+1,    h-2,  theme.frame_vline)

    @contextmanager
    def attr(self, attr_desc):
        """Set attribute for block of commands, then restore previous attribute."""
        original_attr = self.getattr()
        try:
            self.setattr(attr_desc)
            yield
        finally:
            self.setattr(original_attr)


class BufferClippingMixin:

    @property
    def clip_rect(self):
        try:
            return self._clip_rect
        except AttributeError:
            self.clip_rect = None
            return self._clip_rect

    @clip_rect.setter
    def clip_rect(self, value):
        if value:
            self._clip_rect = value
        else:
            self._clip_rect = Rect(0, 0, self.size.w, self.size.h)

    def testxy(self, x, y):
        """Test if x/y coords are inside current clipping region."""
        return (x, y) in self.clip_rect

    @contextmanager
    def clip(self, x, y, w, h):
        """Update clipping region as intersection of original and new region.

        Restore original region before returning.

        """
        original_rect = self.clip_rect
        try:
            self.clip_rect = original_rect.intersect(Rect(x, y, w, h))
            yield
        finally:
            self.clip_rect = original_rect


class Buffer(BufferOperationsMixin, BufferClippingMixin):

    """Rectangular character buffer.

    Retains characters and colors drawn into it
    until resized. Contents are reset when resized.

    """

    def __init__(self, w=0, h=0):
        self._size = Size(w, h)
        self._data = None
        self._attr_descs = ['default']
        self._attr_map = {'default': 0}
        self._current_attr = 0
        self.clear()
        self.sig_resized = Signal()

    @property
    def size(self):
        return self._size.readonly()

    def resize(self, w, h):
        """Resize buffer."""
        self._size.update(w, h)
        self.clear()
        self.sig_resized(w, h)

    def clear(self):
        """Reset buffer data."""
        self._data = [Cell() for _i in range(self._size.w * self._size.h)]
        self.clip_rect = None

    def get(self, x, y):
        """Get cell data at `x`, `y` coords.

        Returns tuple (char, attr_desc).

        """
        cell = self._data[y * self._size.w + x]
        return cell.char, self._attr_descs[cell.attr]

    def getattr(self):
        """Get current attribute."""
        return self._attr_descs[self._current_attr]

    def setattr(self, attr_desc):
        """Set attribute to be used for subsequent draw operations."""
        if attr_desc in self._attr_map:
            attr = self._attr_map[attr_desc]
        else:
            attr = len(self._attr_descs)
            self._attr_descs.append(attr_desc)
            self._attr_map[attr_desc] = attr
        self._current_attr = attr

    def putch(self, x, y, ch):
        """Set character on xy coords to c."""
        if self.testxy(x, y):
            self._data[y * self._size.w + x].set(ch, self._current_attr)


class ProxyBuffer(BufferOperationsMixin, BufferClippingMixin):

    """Special buffer which proxies the operations
    to another buffer or buffer-like class."""

    def __init__(self, target):
        self.target = target
        self.sig_resized = Signal()

    @property
    def size(self):
        return self.target.size

    def resize(self, w, h):
        """Resize buffer."""
        self.target.resize(w, h)
        self.sig_resized(w, h)

    def clear(self):
        """Reset buffer data."""
        self.target.clear()
        self.clip_rect = None

    def get(self, x, y):
        """Get cell data at `x`, `y` coords."""
        return self.target.get(x, y)

    def getattr(self):
        """Get current attribute."""
        return self.target.getattr()

    def setattr(self, attr_desc):
        """Set attribute to be used for subsequent draw operations."""
        self.target.setattr(attr_desc)

    def putch(self, x, y, ch):
        """Set character on xy coords to c."""
        if self.testxy(x, y):
            self.target.putch(x, y, ch)

